// SPDX-License-Identifier: MIT
pragma solidity 0.6.2;

library SafeMath {
    
    function add(uint256 a, uint256 b) internal pure returns (uint256) {
        uint256 c = a + b;
        require(c >= a, "SafeMath: addition overflow");

        return c;
    }
    
    function sub(uint256 a, uint256 b) internal pure returns (uint256) {
        return sub(a, b, "SafeMath: subtraction overflow");
    }

    function sub(uint256 a, uint256 b, string memory errorMessage) internal pure returns (uint256) {
        require(b <= a, errorMessage);
        uint256 c = a - b;

        return c;
    }

    function mul(uint256 a, uint256 b) internal pure returns (uint256) {
        if (a == 0) {
            return 0;
        }

        uint256 c = a * b;
        require(c / a == b, "SafeMath: multiplication overflow");

        return c;
    }

    function div(uint256 a, uint256 b) internal pure returns (uint256) {
        return div(a, b, "SafeMath: division by zero");
    }

    function div(uint256 a, uint256 b, string memory errorMessage) internal pure returns (uint256) {
        require(b > 0, errorMessage);
        uint256 c = a / b;

        return c;
    }

    function mod(uint256 a, uint256 b) internal pure returns (uint256) {
        return mod(a, b, "SafeMath: modulo by zero");
    }

    function mod(uint256 a, uint256 b, string memory errorMessage) internal pure returns (uint256) {
        require(b != 0, errorMessage);
        return a % b;
    }
}

contract Ownable {
    address internal _owner;

    event OwnershipTransferred(address indexed previousOwner, address indexed newOwner);

    /**
     * @dev Initializes the contract setting the deployer as the initial owner.
     */
    constructor () public{
        _owner = msg.sender;
        emit OwnershipTransferred(address(0), _owner);
    }

    /**
     * @dev Returns the address of the current owner.
     */
    function owner() public view returns (address) {
        return _owner;
    }

    /**
     * @dev Throws if called by any account other than the owner.
     */
    modifier onlyOwner() {
        require(_owner == msg.sender, "Ownable: caller is not the owner");
        _;
    }

    /**
     * @dev Transfers ownership of the contract to a new account (`newOwner`).
     * Can only be called by the current owner.
     */
    function transferOwnership(address newOwner) public onlyOwner {
        require(newOwner != address(0), "Ownable: new owner is the zero address");
        emit OwnershipTransferred(_owner, newOwner);
        _owner = newOwner;
    }
    
}


library ECDSA1 {

  /**
   * @dev Recover signer address from a message by using their signature
   * @param hash bytes32 message, the hash is the signed message. What is recovered is the signer address.
   * @param signature bytes signature, the signature is generated using web3.eth.sign()
   */
  function recover(bytes32 hash, bytes memory signature)
    internal
    pure
    returns (address)
  {
    bytes32 r;
    bytes32 s;
    uint8 v;

    // Check the signature length
    if (signature.length != 65) {
      return (address(0));
    }

    // Divide the signature in r, s and v variables with inline assembly.
    assembly {
      r := mload(add(signature, 0x20))
      s := mload(add(signature, 0x40))
      v := byte(0, mload(add(signature, 0x60)))
    }

    // Version of signature should be 27 or 28, but 0 and 1 are also possible versions
    if (v < 27) {
      v += 27;
    }

    // If the version is correct return the signer address
    if (v != 27 && v != 28) {
      return (address(0));
    } else {
      // solium-disable-next-line arg-overflow
      return ecrecover(hash, v, r, s);
    }
  }

  /**
    * toEthSignedMessageHash
    * @dev prefix a bytes32 value with "\x19Ethereum Signed Message:"
    * and hash the result
    */
  function toSignedMessageHash(bytes32 hash)
    internal
    pure
    returns (bytes32)
  {
    return keccak256(
      abi.encodePacked("\x19Ethereum Signed Message:\n32", hash)
    );
  }
}

interface  IBEP2011 {
    function transfer(address to, uint256 value) external returns (bool);

    function mint(address to, uint256 value) external ;

    function approve(address spender, uint256 value) external returns (bool);

    function transferFrom(address from, address to, uint256 value) external returns (bool);

    function totalSupply() external view returns (uint256);

    function balanceOf(address who) external view returns (uint256);

    function allowance(address owner, address spender) external view returns (uint256);
    
    function burn(address account, uint256 value) external returns(bool);

    event Transfer(address indexed from, address indexed to, uint256 value);

    event Approval(address indexed owner, address indexed spender, uint256 value);
}

interface priceAsset {
    
    function getUSDPPriceForLUCAX() external view  returns(uint);
    
    function getUSDPPriceForCALETAS() external view returns (uint);
}

interface Treasury {
    function receiveTeasury( uint _value) external returns (bool);
}


contract CalestaTest {
    using SafeMath for uint;

    string public constant name = 'CALETASTESTLUCAX LP';
    string public constant symbol = 'CALETASTEST-LUCAX';
    uint8 public constant decimals = 18;
    uint  public totalSupply;
    mapping(address => uint) public balanceOf;
    mapping(address => mapping(address => uint)) public allowance;

    bytes32 public DOMAIN_SEPARATOR;
    // keccak256("Permit(address owner,address spender,uint256 value,uint256 nonce,uint256 deadline)");
    bytes32 public constant PERMIT_TYPEHASH = 0x6e71edae12b1b97f4d1f60370fef10105fa2faae0126114a169c64845d6126c9;
    mapping(address => uint) public nonces;

    event Approval(address indexed owner, address indexed spender, uint value);
    event Transfer(address indexed from, address indexed to, uint value);

    constructor() public {
        uint chainId;
        assembly {
            chainId := chainid()
        }
        DOMAIN_SEPARATOR = keccak256(
            abi.encode(
                keccak256('EIP712Domain(string name,string version,uint256 chainId,address verifyingContract)'),
                keccak256(bytes(name)),
                keccak256(bytes('1')),
                chainId,
                address(this)
            )
        );
    }

    function _mint(address to, uint value)  internal {
        totalSupply = totalSupply.add(value);
        balanceOf[to] = balanceOf[to].add(value);
        emit Transfer(address(0), to, value);
    }

    function _burn(address from, uint value) internal {
        balanceOf[from] = balanceOf[from].sub(value);
        totalSupply = totalSupply.sub(value);
        emit Transfer(from, address(0), value);
    }

    function _approve(address owner, address spender, uint value) private {
        allowance[owner][spender] = value;
        emit Approval(owner, spender, value);
    }

    function _transfer(address from, address to, uint value) private {
        balanceOf[from] = balanceOf[from].sub(value);
        balanceOf[to] = balanceOf[to].add(value);
        emit Transfer(from, to, value);
    }

    function approve(address spender, uint value) external returns (bool) {
        _approve(msg.sender, spender, value);
        return true;
    }

    function transfer(address to, uint value) external returns (bool) {
        _transfer(msg.sender, to, value);
        return true;
    }

    function transferFrom(address from, address to, uint value) external returns (bool) {
        if (allowance[from][msg.sender] != uint(-1)) {
            allowance[from][msg.sender] = allowance[from][msg.sender].sub(value);
        }
        _transfer(from, to, value);
        return true;
    }

    function permit(address owner, address spender, uint value, uint deadline, uint8 v, bytes32 r, bytes32 s) external {
        require(deadline >= block.timestamp, 'ERNEswapV2: EXPIRED');
        bytes32 digest = keccak256(
            abi.encodePacked(
                '\x19\x01',
                DOMAIN_SEPARATOR,
                keccak256(abi.encode(PERMIT_TYPEHASH, owner, spender, value, nonces[owner]++, deadline))
            )
        );
        address recoveredAddress = ecrecover(digest, v, r, s);
        require(recoveredAddress != address(0) && recoveredAddress == owner, 'ERNEswapV2: INVALID_SIGNATURE');
        _approve(owner, spender, value);
    }
}

contract signatureVerification{
    
    // keccak256("submit(uint256 owner, uint256 amount, uint256 week, uint256 nonce, uint deadline, bytes memory signature)");
    bytes32 public constant SIGNATURE_PERMIT_TYPEHASH = keccak256("submit(uint256 owner, uint256 amount, uint256 week, uint256 nonce, uint deadline, bytes memory signature)");
    uint public chainId;
    
    using ECDSA1 for bytes32;
    
    constructor() public {
        uint _chainId;
        assembly {
            _chainId := chainid()
        }
        
        chainId = _chainId;
    }
    
    mapping(address => mapping(uint256 => bool)) public seenNonces;
    
    function hashCreation() internal pure returns (bytes32  messageHash){
        bytes32 hash = keccak256(abi.encodePacked());
        messageHash = hash.toSignedMessageHash();
    }
    
    function submit(address owner, uint256 amount, uint256 week, uint256 nonce, uint deadline, bytes memory signature) public {
      // This recreates the message hash that was signed on the client.
      bytes32 hash = keccak256(abi.encodePacked(SIGNATURE_PERMIT_TYPEHASH,owner, amount, week, nonce, chainId, deadline));
      bytes32 messageHash = hash.toSignedMessageHash();
    
      // Verify that the message's signer is the owner of the order
      address signer = messageHash.recover(signature);
      require(signer == owner);
    
      require(!seenNonces[signer][nonce]);
      seenNonces[signer][nonce] = true;
    }
}

contract Pausable is Ownable {
  event Paused();
  event Unpaused();

  bool public _paused = false;


  /**
   * @return true if the contract is paused, false otherwise.
   */
  function paused() public view returns(bool) {
    return _paused;
  }

  /**
   * @dev Modifier to make a function callable only when the contract is not paused.
   */
  modifier whenNotPaused() {
    require(!_paused);
    _;
  }

  /**
   * @dev Modifier to make a function callable only when the contract is paused.
   */
  modifier whenPaused() {
    require(_paused);
    _;
  }

  /**
   * @dev called by the owner to pause, triggers stopped state
   */
  function pause() public onlyOwner  whenNotPaused {
    _paused = true;
    emit Paused();
  }

  /**
   * @dev called by the owner to unpause, returns to normal state
   */
  function unpause() public onlyOwner  whenPaused {
    _paused = false;
    emit Unpaused();
  }
}

contract LucaxStake is Pausable, CalestaTest, signatureVerification {
    
    using SafeMath for uint;
    
    enum Week { Sunday, Monday, Tuesday, Wednesday, Thursday, Friday, Saturday } // days in week
    
    Week public week;
    
    struct stakers{ // user: user info
        uint identity;
        uint refID;
        uint totalStakes;
        uint totalRefStakes;
        bool exist;
    }
    
    struct stake{ // stake: stake info
        uint identity;
        uint sType;
        uint stakeAmount;
        uint mined;
        uint stakeTime;
        uint activation;
        uint currMonth;
        uint lastWithdrawalWeek;
        uint lastWithdraw;
        uint8 currWeek;
        bool isStakeExist;
    }
    
    struct sRewardsReduction{    // sRewards: Mining reduction
        uint fromDay;
        uint endDay;
        mapping( uint => uint) sType;
    }
    
    address payable public EMERGENCY_LIQUIDITY_VAUL;
    address payable public DEVFoundation;
    address payable public TGB;
    
    uint public total_stake_pool;
    
    uint public currVault = 0;
    uint public stakeID;
    uint public ACTIVATE_STAKE = 48 minutes; // ACTIVATE_STAKE: 48 hours (stake active state for mining after 48 hours).
    uint public stakeFee = 12; // stakeFee: Mining fee 12% deduction on every Daily Roi withdrawal.
    
    uint public stakeDays = 17304 minutes; // stakeDays: staking days for active stake is 721 days.
    uint public reductionMonth = 2160 minutes; // reductionMonth: reduction happens every 90 days.
    uint public reductionCurrMnth = 0; // reductionCurrMnth: reduction month start from march 2021
    uint public REDUCTION_LIMIT = 9; // REDUCTION_LIMIT: reduction limit
    
    uint public penalty_period  = 2160 minutes; // penalty_period: penality period 
    uint public penaltyLiquidity;
    
    uint public stakeActivation = now;

    uint[5][] public miningTable = [
        [11e18,12e18,13e18,stakeActivation,stakeActivation.add(2160 minutes)],[8e18,9e18,10e18,stakeActivation.add(2184 minutes),stakeActivation.add(4320 minutes)],
        [7e18,8e18,9e18,stakeActivation.add(4344 minutes),stakeActivation.add(6480 minutes)],[6e18,7e18,8e18,stakeActivation.add(6504 minutes),stakeActivation.add(8640 minutes)],
        [4e18,5e18,6e18,stakeActivation.add(8664 minutes),stakeActivation.add(10800 minutes)],[3e18,4e18,5e18,stakeActivation.add(10824 minutes),stakeActivation.add(12960 minutes)],
        [2e18,3e18,4e18,stakeActivation.add(12984 minutes),stakeActivation.add(15120 minutes)],[1e18,2e18,3e18,stakeActivation.add(15144 minutes),stakeActivation.add(17280 minutes)],
        [0.8e18,1e18,2e18,stakeActivation.add(17304 minutes),stakeActivation.add(19440 minutes)]
    ];
    
    
    mapping(address => stakers) public staker;
    mapping(uint => address) public stakerById;
    mapping(uint => stake) public stakes;
    mapping(uint => sRewardsReduction) public sReward;
    mapping(uint => uint) public lucaxStake;
    mapping(address => uint[]) public _totMondays;
    mapping(address => uint) public activeLUCAXPlan500;
    
    event stakeEvent( address _staker, uint indexed _stakerID, uint indexed _stakeID, uint _stakeAmount, uint _time);
    event stakeRewardsEvent( address _staker, uint indexed _stakerID, uint indexed _stakeID, uint _stakeReward, uint _time);
    event promoRewardEvent(address _staker, uint indexed _stakerID, uint _promoReward, uint _time);
    event unstakeEvent( address _staker, uint indexed _stakerID, uint indexed _stakeID, uint _unstakeAmount, uint _penalty, uint _time);
    event transferStakeOwnerShipEvent( address _staker, uint indexed _stakerID, address _trnsStaker, uint indexed _trnsStakerID, uint indexed _stakeID, uint _time);
    event activeStakeEvent( address _staker, uint indexed _stakerID, uint _stakeTime, uint _time);
    
    IBEP2011 public LUCAX;
    
    priceAsset public ASSET;
    
    Treasury public TREASURY;
    
    constructor( address _lucax, priceAsset _Asset, address payable _treasury, address payable _EMERGENCY_LIQUIDITY_VAUL, address payable _DEVFoundation) public {
        initiate();
        
        lucaxStake[0] = 5e18;
        lucaxStake[1] = 50e18;
        lucaxStake[2] = 200e18;
        lucaxStake[3] = 500e18;
        
        LUCAX = IBEP2011(_lucax);
        ASSET = _Asset;
        TREASURY = Treasury(_treasury);
        
        EMERGENCY_LIQUIDITY_VAUL = _EMERGENCY_LIQUIDITY_VAUL;
        DEVFoundation = _DEVFoundation;
        TGB = _treasury;
    }
    
    function initiate() internal {
        for(uint i;i< miningTable.length;i++){
            sReward[i].sType[0] = miningTable[i][0];
            sReward[i].sType[1] = miningTable[i][1];
            sReward[i].sType[2] = miningTable[i][2];
            sReward[i].fromDay = miningTable[i][3];
            sReward[i].endDay = miningTable[i][4];
        }
    }
    
//function stakeLUCAX( uint _typ, uint _refID, uint _lucax, uint8 _week, uint256 _nonce, uint _deadline, bytes memory _signature) public whenNotPaused returns (bool){ 
    function stakeLUCAX( uint _typ, uint _refID, uint _lucax, uint8 _week) public whenNotPaused returns (bool){ 
        require((_typ >= 0) && (_typ <= 2), '_typ: stake plan must be 0 to 2');
        require((lucaxStake[_typ] <= _lucax) && (lucaxStake[_typ.add(1)] >= _lucax), '_lucax: Invalid _lucax');
        
        require((_week >= uint(Week.Sunday)) && (_week <= uint(Week.Saturday)), '_week: _week should between sunday to saturday');
        
	    if( currVault == 0 ) { require(_refID == 0, '_refID: Invalid referrer ID'); } else { require(_refID >=0 && _refID <= currVault, '_refID: Invalid referrer ID'); }

       // submit( msg.sender, _lucax, _week, _nonce, _deadline, _signature);
        
        require(LUCAX.balanceOf(msg.sender) >= _lucax,"insufficient balance");
        
        if(!staker[msg.sender].exist){ createVault( _refID); }
        
        _createStake( msg.sender, _typ, _lucax, _week);
        
        if((block.timestamp >= sReward[0].fromDay) && (block.timestamp <= sReward[0].endDay)){ if(staker[msg.sender].refID > 0) { referralBonus( stakerById[staker[msg.sender].refID] , _lucax); } staker[msg.sender].totalRefStakes = staker[msg.sender].totalRefStakes.add(_lucax); }
        
        require(LUCAX.burn( msg.sender, _lucax), "burn failed"); // burn staker token
        total_stake_pool = total_stake_pool.add(_lucax);

        uint _mintable = (((ASSET.getUSDPPriceForCALETAS().mul(1e18)).div(ASSET.getUSDPPriceForLUCAX())).mul(_lucax)).div(1e18);
        
        _mint( msg.sender, _mintable);
        
        emit stakeEvent( msg.sender, staker[msg.sender].identity, stakeID, _lucax, block.timestamp);
        
        return true;
    }
    
    function getStakeDROI( uint _stakeID) public returns (bool){
        require(block.timestamp > sReward[0].endDay, 'getStakeDROI: stakes can withdraw their reward after promotion period');
        require(staker[msg.sender].exist, 'Exist: staker does not exist');
        require(stakes[_stakeID].identity == staker[msg.sender].identity, 'Access denied: invalid access');
        require(block.timestamp > stakes[_stakeID].stakeTime.add(stakes[_stakeID].activation), 'Active stake: staker can withdraw thier reward after stake activation');
        require(stakes[_stakeID].isStakeExist, 'isStakeExist: stake does not exist');
        
        _updateReduction();
        
        (uint _DROI, uint _lastWithdrawalWeek, uint _currWeek, uint _currMonth) = _calculateDROI( _stakeID);

       
        
        if(_DROI == 0) { return false; }
        
        _stakeFeeDistribution( _DROI); // yield mining fee distribution
        
        _DROI = _DROI.sub(_calculateFee( _DROI, stakeFee));
        
        if(_DROI > 0){ _sMint(msg.sender, _DROI); }
        
        stakes[_stakeID].currWeek = uint8(_currWeek);
        stakes[_stakeID].currMonth = _currMonth;
        stakes[_stakeID].mined = _DROI;
        stakes[_stakeID].lastWithdrawalWeek = _lastWithdrawalWeek;
        
        emit stakeRewardsEvent( msg.sender, staker[msg.sender].identity, _stakeID, _DROI, block.timestamp);
        
        return true;   
    }
    
    function unStake( uint _stakeID) public returns (bool){
        require(staker[msg.sender].exist, 'Exist: staker does not exist');
        require(stakes[_stakeID].identity == staker[msg.sender].identity, 'Access denied: invalid access');
        require(stakes[_stakeID].isStakeExist, 'isStakeExist: stake does not exist');
        
        uint _burnable = (((ASSET.getUSDPPriceForCALETAS().mul(1e18)).div(ASSET.getUSDPPriceForLUCAX())).mul(stakes[_stakeID].stakeAmount)).div(1e18);
        
        require(balanceOf[msg.sender] >= _burnable, 'balanceOf: insufficient caletas');
        
        uint _penalty;
        
        if(block.timestamp >= stakes[_stakeID].stakeTime.add(stakes[_stakeID].activation)){ _penalty = _calculatePenalty( _stakeID); }
        
        uint _reMint = (((ASSET.getUSDPPriceForCALETAS().mul(1e18)).div(ASSET.getUSDPPriceForLUCAX())).mul(_burnable)).div(1e18);
        
        if(_penalty > 0){
            _penalty = _reMint.mul(_penalty).div(100);   
            penaltyLiquidity = penaltyLiquidity.add(_penalty);
            _reMint = _reMint.sub(_penalty);
        }
        
        _burn(msg.sender, _burnable); // caletas token burn
        
        uint penaltyPropotion = _computeStakePropotion( _stakeID); // calculate penalty propotion
        
        penaltyPropotion = _computePenaltyPropotion( penaltyPropotion);
        
        _sMint(msg.sender, penaltyPropotion); // mint penalty propotion
        penaltyLiquidity = penaltyLiquidity.sub(penaltyPropotion);
        
        _sMint(msg.sender, _reMint);
        (total_stake_pool >= _reMint) ? total_stake_pool = total_stake_pool.sub(_reMint) : total_stake_pool = 0;
        
        if(stakes[_stakeID].sType == 0){ activeLUCAXPlan500[_owner]--; }

        stakes[_stakeID].isStakeExist = false;
        
        emit unstakeEvent( msg.sender, staker[msg.sender].identity, _stakeID, _reMint, _penalty, block.timestamp);
        
        return true;
    }
    
    function referralBonus( address _sponser,uint _contribution) private returns (bool){
        uint _percent;
        
        uint USDP = ASSET.getUSDPPriceForLUCAX().mul(staker[_sponser].totalRefStakes);
        
        (USDP < 10e18) ? _percent = 5 : _percent = 10;
        
        uint _promoReward = _contribution.mul(_percent).div(100);
        
        if(_promoReward > 0){ _sMint(_sponser, _promoReward); emit promoRewardEvent( _sponser, staker[_sponser].identity, _promoReward, block.timestamp); }
        
        return true;
    }
    
    function transferStakeOwnerShip( uint _stakeID, uint _transferID) public returns (bool){
        require(staker[msg.sender].exist, 'Exist: staker does not exist');
        require(stakes[_stakeID].identity == staker[msg.sender].identity, 'Access denied: invalid access');
        require((_transferID > 0) && (stakes[_stakeID].identity != _transferID) && (_transferID <= currVault), '_transferID: transfer stake id does not exist');
        require(block.timestamp < stakes[_stakeID].stakeTime.add(stakes[_stakeID].activation), 'ACTIVATE_STAKE: cannot make transfer, stake activated for mining');
        require(stakes[_stakeID].isStakeExist, "transferStakeOwnerShip :: Stake is not exist");
        
        stakes[_stakeID].identity = _transferID;
        
        emit transferStakeOwnerShipEvent( msg.sender, staker[msg.sender].identity, stakerById[_transferID], _transferID, _stakeID, block.timestamp);
        
        return true;
    }
    
    function activateStake( uint _stakeID, uint _stakActivation) public  returns (bool){
        require(staker[msg.sender].exist, 'Exist: staker does not exist');
        require(_stakActivation > 0, "activateStake:: _stakActivation: must be greater than zero");
        require(stakes[_stakeID].identity == staker[msg.sender].identity, 'Access denied: invalid access');
        require(_stakActivation > stakes[_stakeID].activation, "_stakeActivation must be greater than activation");
        require(stakes[_stakeID].stakeTime.add(stakes[_stakeID].activation) < block.timestamp, 'ACTIVATE_STAKE: stake activate');
        
        stakes[_stakeID].activation = _stakActivation;
        stakes[_stakeID].lastWithdrawalWeek = stakes[_stakeID].stakeTime.add(stakes[_stakeID].activation);
        
        emit activeStakeEvent( msg.sender, _stakeID, _stakActivation, block.timestamp);
        
        return true;
    }
    
    function setASSET( address _Asset) public onlyOwner returns (bool){
        require(_Asset != address(0), "_Asset: ASSET contract address must not be zero");
        
        ASSET = priceAsset(_Asset);
        return true;
    }
    
    function rewardReductionInfo( uint _mnth) public view returns( uint _s0,uint _s1,uint _s2 ){
        return(sReward[_mnth].sType[0],sReward[_mnth].sType[1],sReward[_mnth].sType[2]);
    }
    
    // --- private ---
    
    function _stakeFeeDistribution( uint _amount) private returns (bool){
       uint _EMGL_DEV;
       uint _TGB;
       
       if(_calculateFee( _amount, 2) > 0) { _EMGL_DEV = _calculateFee( _amount, 2); }
       if(_calculateFee( _amount, 8) > 0) { _TGB = _calculateFee( _amount, 8); }
       
       _sMint(EMERGENCY_LIQUIDITY_VAUL,_EMGL_DEV);
       _sMint(DEVFoundation,_EMGL_DEV);
       _sMint(TGB,_TGB);
       
       require(TREASURY.receiveTeasury(_TGB), "_stakeFeeDistribution:: Treasury fund tracker failed");
       
       return true;
    }
    
    function _sMint(address _miner, uint256 value) private {
        if(LUCAX.balanceOf(address(this)) >= value ){ LUCAX.transfer(msg.sender,value); }
        else { LUCAX.mint(_miner,value); }
    }
    
    function _updateReduction() public returns (bool){
        if(sReward[reductionCurrMnth].endDay < block.timestamp){
            uint durationExceed = block.timestamp.sub(sReward[reductionCurrMnth].endDay).div(reductionMonth);
            if(durationExceed ==0 ){ return false; }
            if(reductionCurrMnth.add(durationExceed) > REDUCTION_LIMIT){
                uint remainDuration = REDUCTION_LIMIT.sub(reductionCurrMnth);
                if(remainDuration == 0){ durationExceed = REDUCTION_LIMIT; }
                else { durationExceed = remainDuration;  }
                
                reductionCurrMnth = reductionCurrMnth.add(durationExceed);
                return true;
            }
            else { reductionCurrMnth = reductionCurrMnth.add(durationExceed); return true; }
        }
        return false;
    }
    
    function _calculateDROI( uint _stakeID) public view returns (uint tot_payout, uint _lastweekWithdraawal, uint _currentWeek, uint _currentMonth){
        
        uint currTime = block.timestamp;
        
        uint stakeid = _stakeID;
        
        if(currTime > stakes[stakeid].stakeTime.add(stakeDays)){ currTime = stakes[stakeid].stakeTime.add(stakeDays); }
        
        _lastweekWithdraawal = stakes[stakeid].lastWithdrawalWeek;
        
        _currentMonth = stakes[stakeid].currMonth;
        
        _currentWeek = stakes[stakeid].currWeek;
        
        uint _days = currTime.sub(stakes[stakeid].lastWithdrawalWeek).div(24 minutes);
        
        uint[3] memory lastWeek;
        lastWeek[0] = stakes[stakeid].currWeek;
        lastWeek[1] = 1;
        
        if(_days > 0) {
            for( uint i=1;i<=_days;i++){
                uint _dayOnWeek = lastWeek[0].add(lastWeek[1]);
                lastWeek[2] = lastWeek[2].add(1);
                
                if(_dayOnWeek == uint(Week.Monday)){
                    
                    for(uint j=1;j<=lastWeek[2];j++){                                                
                        if(_lastweekWithdraawal > sReward[_currentMonth].endDay)
                            if(_currentMonth.add(1) < REDUCTION_LIMIT){ _currentMonth = _currentMonth.add(1); }
                        
                        tot_payout = tot_payout.add((stakes[stakeid].stakeAmount.mul(sReward[_currentMonth].sType[stakes[stakeid].sType]).div(100e18)));

                        _lastweekWithdraawal = _lastweekWithdraawal.add(24 minutes);
                    }
                    
                    _currentWeek = 1;
                    lastWeek[0] = 1;
                    lastWeek[2] = 0;
                }
                else if(_dayOnWeek > uint(Week.Saturday)){
                    lastWeek[1] = 1;
                    lastWeek[0] = 0;
                }
                else
                    lastWeek[1]++;
            }
        }
    }
    
    function _computePenaltyPropotion( uint _propotion) public view returns (uint _penaltyPropotion){
        _penaltyPropotion = penaltyLiquidity.mul(_propotion).div(100e18);
    }
    
    function _computeStakePropotion( uint _stakeID) public view returns (uint _propotion){
        _propotion = stakes[_stakeID].stakeAmount.mul(100e18).div(total_stake_pool);
    }
    
    function createVault( uint _refID) private returns (bool){
        currVault++;
        
        stakers memory _userStruct;
        
        _userStruct = stakers({
            identity: currVault,
            refID: _refID,
            totalStakes: 0,
            totalRefStakes: 0,
            exist: true
        });
        
        staker[msg.sender] = _userStruct;
        
        stakerById[currVault] = msg.sender;
        
        return true;
    }
    
    function _createStake( address _owner, uint _typ, uint _lucax, uint8 _week) private returns (bool){
        stakeID++; 
        stake memory _stake;
        _stake = stake({
            identity: staker[_owner].identity,
            sType: _typ,
            stakeAmount: _lucax,
            mined : 0,
            stakeTime: block.timestamp,
            activation: ACTIVATE_STAKE,
            lastWithdraw: block.timestamp,
            lastWithdrawalWeek: block.timestamp.add(ACTIVATE_STAKE),
            currWeek: _week,
            currMonth: reductionCurrMnth,
            isStakeExist: true
        });
            
        stakes[stakeID] = _stake;
        
        if(_typ == 0){ activeLUCAXPlan500[_owner]++; }
        
        return true;
    }
    
    function _getUSDPCurPriceOfLUCAX( ) public view returns ( uint _value){
        _value = uint(60000e18).mul(1e18).div(ASSET.getUSDPPriceForLUCAX());
    }
    
    function _calculatePenalty( uint _stakeID) public view returns (uint){
        uint stake_time = stakes[_stakeID].stakeTime.add(ACTIVATE_STAKE);
        uint percent = 90;
        uint reductPercent = 10;
        
        for(uint i=1;i<=8;i++){
            if((block.timestamp > stake_time) && (block.timestamp <= stake_time.add(penalty_period.mul(i)))) { return percent.sub(reductPercent.mul(i)); }
        }
        
        return 0;
    }
    
    function _calculateFee( uint _amount, uint _percentage) public pure returns (uint _fee){
        _fee = _amount.mul(_percentage).div(100);
    }
}